/*

		File List Utility

*/

#include "screensv.h"

struct fileListItem **fileList = NULL;	// File list
int fileListLength = 0;			// Number of files in list
static int fileListMax = 0;		// Maximum files in current list

struct fileListItem **soundFileList = NULL;	// Sound file list
int soundFileListLength = 0;	// Number of files in list
static int soundFileListMax = 0;// Maximum files in current list

int fileListStatus = 0;			// Status from building file list
#ifdef INTERNET_FILES
bool coInitialised = FALSE;		// Have we initialised COM ?
#endif

#define InitialList	100			// Initial list size
#define GrowList	50			// Growth increment

/*	fileListFilter  --  Filter function to determine
						whether to include a file in the
						list.  Passed the WIN32_FIND_DATA,
						it returns 0 if the file is not
						to be included, 1 if the file is an
						image file, and 2 if the file is a
						sound file.  */

static int fileListFilter(WIN32_FIND_DATA *fd)
{
	int i;
	/*	A leading "." denotes an image file, "!" a wave audio file,
		"#" is a MIDI file, '+' an MP3 file, '-' an Internet URL shortcut. */

	char *extensions[] = {
		".jpg",
		".gif",
		".bmp",
		".png",
		".jif",
		".jpeg",
		"!wav",
		"#mid",
		"#rmi",
		"+mp3",
#ifdef INTERNET_FILES
		"-url",
#endif
		NULL
	};

	//	Skip directories and zero-length files

	if ((fd->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ||
		((fd->nFileSizeHigh | fd->nFileSizeLow) == 0)) {
		return FALSE;
	}

	for (i = 0; extensions[i] != NULL; i++) {
		char extt[10];

		strcpy(extt, extensions[i]);
		extt[0] = '.';
		if (_stricmp(extt,
				fd->cFileName +
					(strlen(fd->cFileName) - strlen(extensions[i]))) == 0) {
			return (*extensions[i] == '.') ? 1 : ((*extensions[i] == '!') ? 2 :
					((*extensions[i] == '#') ? 3 :
					  ((*extensions[i] == '+') ? 4 : 
						5)));
		}
	}
	return FALSE;
}

//	file ListDestroy  --  Destroy a file list, releasing all storage

void fileListDestroy(void)
{
	int i;

	if (fileList != NULL) {
		for (i = 0; i < fileListLength; i++) {
			if (fileList[i]->iURL != NULL) {
				free(fileList[i]->iURL);
			}
			free(fileList[i]);
		}
		free(fileList);	
	}
	fileList = NULL;
	fileListLength = fileListMax = 0;

	if (soundFileList != NULL) {
		for (i = 0; i < soundFileListLength; i++) {
			if (soundFileList[i]->iURL != NULL) {
				free(soundFileList[i]->iURL);
			}
			free(soundFileList[i]);
		}
		free(soundFileList);	
	}
	soundFileList = NULL;
	soundFileListLength = soundFileListMax = 0;
}

//	getInternetShortcutURL  --  Resolve an Internet shortcut into the URL

#ifdef INTERNET_FILES

static BOOL getInternetShortcutURL(const char *path, const char *filename, char *url, unsigned int urlLen)
{
	IMalloc *pMalloc;
	IUniformResourceLocator *pUrl;
	HRESULT hr;
	char *pfn;
		
	if (!coInitialised) {
		CoInitialize(NULL);
		coInitialised = TRUE;
	}
		
	hr = SHGetMalloc(&pMalloc);
	if (!SUCCEEDED(hr)) {
		return FALSE;
	}
	
	hr = CoCreateInstance(CLSID_InternetShortcut, NULL, CLSCTX_INPROC_SERVER,
						  IID_IUniformResourceLocator, (LPVOID *) &pUrl);
	if (!SUCCEEDED(hr)) {
		pMalloc->Release();
		return FALSE;
	}
	
	IPersistFile *ppf;
	
	hr = pUrl->QueryInterface(IID_IPersistFile, (LPVOID *) &ppf);
	if (!SUCCEEDED(hr)) {
		pUrl->Release();
		pMalloc->Release();
		return FALSE;
	}

	WCHAR wszPath[_MAX_PATH];
	
	pfn = (char *) malloc(strlen(path) + strlen(filename) + 2);
	if (pfn == NULL) {
		pUrl->Release();
		pMalloc->Release();
		return FALSE;
	}
	strcpy(pfn, path);
	*strrchr(pfn, '/') = EOS;				// Trust me, it's there
	strcat(pfn, "\\");
	strcat(pfn, filename);
	MultiByteToWideChar(CP_OEMCP, 0, pfn, -1, wszPath, _MAX_PATH);
	free(pfn);
	hr = ppf->Load(wszPath, STGM_READ);
	
	BOOL stat = FALSE;
	if (SUCCEEDED(hr)) {
		LPSTR pszUrl;
		
		hr = pUrl->GetURL(&pszUrl);
		if (SUCCEEDED(hr)) {

			lstrcpyn(url, pszUrl, urlLen);
			pMalloc->Free(pszUrl);
			stat = TRUE;
		}
	}
	
	ppf->Release();
	pUrl->Release();
	pMalloc->Release();
	return stat;
}
#endif

//	fileListBuild  --  Scan a directory and build a file list

int fileListBuild(const char *path)
{
	HANDLE sitem;
	WIN32_FIND_DATA fd;
#ifdef INTERNET_FILES
	DWORD icflags;
	BOOL ignoreURLlinks = ignoreInternetShortcuts ||
			(!InternetGetConnectedState(&icflags, 0));
#endif

	fileListStatus = 0;
	fileListDestroy();
	fileList = (struct fileListItem **) malloc(InitialList * sizeof(struct fileListItem *));
	if (fileList == NULL) {
		return 0;
	}
	fileListMax = InitialList;

	sitem = FindFirstFile(path, &fd);

	if (sitem == INVALID_HANDLE_VALUE) {
		fileListDestroy();
		fileListStatus = FL_BADDIR;				// Invalid directory
		return 0;
	}

	while (TRUE) {
		struct fileListItem *fi;
		int filetype = fileListFilter(&fd);
		char *fURL = NULL;
		
#ifdef INTERNET_FILES
		if ((filetype == 5) && ignoreURLlinks) {
			filetype = 0;
		}
		
		if (filetype == 5) {
			char url[_MAX_PATH];

#ifdef INTERNET_FILES_DEBUG
			char s[132];
			wsprintf(s, "Internet shortcut: (%s)\r\n", fd.cFileName);
			OutputDebugString(s);
#endif
			
			if (getInternetShortcutURL(path, fd.cFileName, url, sizeof url)) {
				WIN32_FIND_DATA afd = fd;
#ifdef INTERNET_FILES_DEBUG
				OutputDebugString("   ...OK.  URL = ");
				OutputDebugString(url);
				OutputDebugString("\r\n");
#endif
				fURL = url;
				
				if ((strlen(afd.cFileName) > 4) &&
					(_stricmp(".url", afd.cFileName + (strlen(afd.cFileName) - 4))== 0)){
					afd.cFileName[strlen(afd.cFileName) - 4] = EOS;
				}
				filetype = fileListFilter(&afd) | 0x80;
				
				/* If fileListFilter can't identify the file type from the name,
				   assume it's an image.  Sound file must be explicitly identified
				   with a name like Kaboom.wav.url. */
				   
				if ((filetype & 0x7F) == 0) {
					filetype |= 1;
				}
			} else {
#ifdef INTERNET_FILES_DEBUG
				OutputDebugString("   ...Error\r\n");
#endif
				filetype = 0;
			}
		}
#endif

		if ((filetype & 0x7F) != 0) {
			if ((filetype & 0x7F) == 1) {
			
				//	This is an image file
				
				if (fileListLength >= fileListMax) {
					struct fileListItem **np = (struct fileListItem **) malloc(
						(fileListMax + GrowList) * sizeof(struct fileListItem *));

					if (np == NULL) {
						fileListDestroy();
						fileListStatus = FL_NOMEM;
						return 0;
					}
					/*	Why not use GlobalReAlloc?  Because it doesn't
						bloody work--typical Microsoft crap.  In any
						case, having been written by Kode Kiddies, it
						probably always copies anyway.  */
					memcpy(np, fileList, fileListMax * sizeof(struct fileListItem *));
					free(fileList);
					fileList = np;
					fileListMax += GrowList;

				}
				fi = (struct fileListItem *) malloc(sizeof(struct fileListItem) + strlen(fd.cFileName) + 1);
				if (fi == NULL) {
					fileListDestroy();
					fileListStatus = FL_NOMEM;
					return 0;
				}
				fi->age = 0;
				fi->iURL = (fURL != NULL) ? _strdup(fURL) : NULL;
				fi->fileType = filetype;
				strcpy(fi->filename, fd.cFileName);
				fileList[fileListLength++] = fi;
			} else {
			
				//	This is a sound file
				
				if (soundMode != SOUND_NEVER) {
					if (soundFileListLength >= soundFileListMax) {
						struct fileListItem **np = (struct fileListItem **) malloc(
							(soundFileListMax + GrowList) * sizeof(struct fileListItem *));

						if (np == NULL) {
							fileListDestroy();
							fileListStatus = FL_NOMEM;
							return 0;
						}
						memcpy(np, soundFileList, soundFileListMax * sizeof(struct fileListItem *));
						free(soundFileList);
						soundFileList = np;
						soundFileListMax += GrowList;

					}
					fi = (struct fileListItem *) malloc(sizeof(struct fileListItem) +
										strlen(fd.cFileName) + 1);
					if (fi == NULL) {
						fileListDestroy();
						fileListStatus = FL_NOMEM;
						return 0;
					}
					fi->age = filetype;
					fi->iURL = (fURL != NULL) ? _strdup(fURL) : NULL;
					fi->fileType = filetype;
					strcpy(fi->filename, fd.cFileName);
					soundFileList[soundFileListLength++] = fi;
				}
			}
		}
		if (FindNextFile(sitem, &fd) == 0) {
			FindClose(sitem);
			break;
		}
	}

	return fileListLength;
}

//	fileListDump  --  Dump file list to a file for debugging

void fileListDump(FILE *df)
{
	int i;

	fprintf(df, "--- Image files ---\n");
	for (i = 0; i < fileListLength; i++) {
		fprintf(df, "%4d: %s\n", fileList[i]->age, fileList[i]->filename);
	}
	fprintf(df, "--- Sound files ---\n");
	for (i = 0; i < soundFileListLength; i++) {
		fprintf(df, "%4d: %s\n", soundFileList[i]->age, soundFileList[i]->filename);
	}
}

//	fileNameCompare  --  Qsort comparison function for sorting file names

static int fileNameCompare(const void *arg1, const void *arg2) {
   /* Compare all of both strings: */
	return _stricmp((*(struct fileListItem **) arg1)->filename,
					(*(struct fileListItem **) arg2)->filename);
}

/*	fileListSort  --  Sort file list into alphabetical order.
					  The above comparison auxiliary is used,
					  which performs case-insensitive comparisons.  */

void fileListSort(void)
{
	if (fileList != NULL) {
		qsort((void *) fileList, (size_t) fileListLength, sizeof(char *), fileNameCompare);
	}
	if (soundFileList != NULL) {
		qsort((void *) soundFileList, (size_t) soundFileListLength, sizeof(char *), fileNameCompare);
	}
}

//	fileListShuffleImages  --  Shuffle image file list into random order

void fileListShuffleImages(void)
{
	if (fileList != NULL) {
		int i;
		int fileListMask = 1;
		while (fileListMask <= fileListLength){
			fileListMask <<= 1;
		}
		fileListMask--;
		
		for (i = 0; i < fileListLength; i++) {
			int j;
			
			do {
				j = (random()>> 5) & fileListMask;
			} while (j >= fileListLength);
			struct fileListItem *fi = fileList[i];
			
			fileList[i] = fileList[j];
			fileList[j] = fi;
		}
	}
}
//	fileListShuffleSounds  --  Shuffle sound file list into random order

void fileListShuffleSounds(void)
{
	if (soundFileList != NULL) {
		int i;
		int soundFileListMask = 1;
		while (soundFileListMask <= soundFileListLength){
			soundFileListMask <<= 1;
		}
		soundFileListMask--;

		for (i = 0; i < soundFileListLength; i++) {
			int j;
			
			do {
				j = (random() >> 5) & soundFileListMask;
			} while (j >= soundFileListLength);
			struct fileListItem *fi = soundFileList[i];
			
			soundFileList[i] = soundFileList[j];
			soundFileList[j] = fi;
		}
	}
}
